@@ -258,6 +258,7 @@ static void fuse_put_super(struct super_block *sb)
 	kill_fasync(&fc->fasync, SIGIO, POLL_IN);
 	wake_up_all(&fc->waitq);
 	wake_up_all(&fc->blocked_waitq);
+	wake_up_all(&fc->reserved_req_waitq);
 	mutex_lock(&fuse_mutex);
 	list_del(&fc->entry);
 	fuse_ctl_remove_conn(fc);
@@ -293,6 +294,11 @@ static int fuse_statfs(struct super_block *sb, struct kstatfs *buf)
 	struct fuse_statfs_out outarg;
 	int err;
 
+	if (!fuse_allow_task(fc, current)) {
+		buf->f_type = FUSE_SUPER_MAGIC;
+		return 0;
+	}
+
 	req = fuse_get_req(fc);
 	if (IS_ERR(req))
 		return PTR_ERR(req);
@@ -469,6 +475,7 @@ static struct fuse_conn *new_conn(void)
 		atomic_set(&fc->count, 1);
 		init_waitqueue_head(&fc->waitq);
 		init_waitqueue_head(&fc->blocked_waitq);
+		init_waitqueue_head(&fc->reserved_req_waitq);
 		INIT_LIST_HEAD(&fc->pending);
 		INIT_LIST_HEAD(&fc->processing);
 		INIT_LIST_HEAD(&fc->io);
@@ -506,27 +513,33 @@ static struct inode *get_root_inode(struct super_block *sb, unsigned mode)
 
 	attr.mode = mode;
 	attr.ino = FUSE_ROOT_ID;
+	attr.nlink = 1;
 	return fuse_iget(sb, 1, 0, &attr);
 }
 #ifndef FUSE_MAINLINE
 #ifdef HAVE_EXPORTFS_H
 #include <linux/exportfs.h>
 #endif
-static struct dentry *fuse_get_dentry(struct super_block *sb, void *vobjp)
+
+struct fuse_inode_handle
+{
+	u64 nodeid;
+	u32 generation;
+};
+
+static struct dentry *fuse_get_dentry(struct super_block *sb,
+				      struct fuse_inode_handle *handle)
 {
-	__u32 *objp = vobjp;
-	unsigned long nodeid = objp[0];
-	__u32 generation = objp[1];
 	struct inode *inode;
 	struct dentry *entry;
 
-	if (nodeid == 0)
+	if (handle->nodeid == 0)
 		return ERR_PTR(-ESTALE);
 
-	inode = ilookup5(sb, nodeid, fuse_inode_eq, &nodeid);
+	inode = ilookup5(sb, handle->nodeid, fuse_inode_eq, &handle->nodeid);
 	if (!inode)
 		return ERR_PTR(-ESTALE);
-	if (inode->i_generation != generation) {
+	if (inode->i_generation != handle->generation) {
 		iput(inode);
 		return ERR_PTR(-ESTALE);
 	}
@@ -536,42 +549,132 @@ static struct dentry *fuse_get_dentry(struct super_block *sb, void *vobjp)
 		iput(inode);
 		return ERR_PTR(-ENOMEM);
 	}
+	entry->d_op = &fuse_dentry_operations;
 
 	return entry;
 }
 
-static int fuse_encode_fh(struct dentry *dentry, __u32 *fh, int *max_len,
-			  int connectable)
+static int fuse_encode_fh(struct dentry *dentry, u32 *fh, int *max_len,
+			   int connectable)
 {
 	struct inode *inode = dentry->d_inode;
 	int len = *max_len;
 	int type = 1;
+	u64 nodeid;
+	u32 generation;
 
-	if (len < 2 || (connectable && len < 4))
-		return 255;
+	if (len < 3 || (connectable && len < 6))
+		return  255;
+
+	nodeid = get_fuse_inode(inode)->nodeid;
+	generation = inode->i_generation;
+
+	len = 3;
+	fh[0] = (u32)(nodeid >> 32);
+	fh[1] = (u32)(nodeid & 0xffffffff);
+	fh[2] = generation;
 
-	len = 2;
-	fh[0] = get_fuse_inode(inode)->nodeid;
-	fh[1] = inode->i_generation;
 	if (connectable && !S_ISDIR(inode->i_mode)) {
 		struct inode *parent;
 
 		spin_lock(&dentry->d_lock);
 		parent = dentry->d_parent->d_inode;
-		fh[2] = get_fuse_inode(parent)->nodeid;
-		fh[3] = parent->i_generation;
+		nodeid = get_fuse_inode(parent)->nodeid;
+		generation = parent->i_generation;
+
+		fh[3] = (u32)(nodeid >> 32);
+		fh[4] = (u32)(nodeid & 0xffffffff);
+		fh[5] = generation;
 		spin_unlock(&dentry->d_lock);
-		len = 4;
+
+		len = 6;
 		type = 2;
 	}
+
 	*max_len = len;
 	return type;
 }
 
-static struct export_operations fuse_export_operations = {
-	.get_dentry	= fuse_get_dentry,
-	.encode_fh      = fuse_encode_fh,
+#ifdef KERNEL_2_6_24_PLUS
+static struct dentry *fuse_fh_to_dentry(struct super_block *sb,
+		struct fid *fid, int fh_len, int fh_type)
+{
+	struct fuse_inode_handle handle;
+
+	if (fh_len < 3 || fh_type > 2)
+		return NULL;
+
+	handle.nodeid = (u64) fid->raw[0] << 32;
+	handle.nodeid |= (u64) fid->raw[1];
+	handle.generation = fid->raw[2];
+	return fuse_get_dentry(sb, &handle);
+}
+
+static struct dentry *fuse_fh_to_parent(struct super_block *sb,
+		struct fid *fid, int fh_len, int fh_type)
+{
+	struct fuse_inode_handle parent;
+
+	if (fh_type != 2 || fh_len < 6)
+		return NULL;
+
+	parent.nodeid = (u64) fid->raw[3] << 32;
+	parent.nodeid |= (u64) fid->raw[4];
+	parent.generation = fid->raw[5];
+	return fuse_get_dentry(sb, &parent);
+}
+
+
+static const struct export_operations fuse_export_operations = {
+	.fh_to_dentry	= fuse_fh_to_dentry,
+	.fh_to_parent	= fuse_fh_to_parent,
+	.encode_fh	= fuse_encode_fh,
 };
+#else
+static struct dentry *fuse_get_dentry_old(struct super_block *sb, void *objp)
+{
+	return fuse_get_dentry(sb, objp);
+}
+
+static struct export_operations fuse_export_operations;
+
+static struct dentry *fuse_decode_fh(struct super_block *sb, u32 *fh,
+			int fh_len, int fileid_type,
+			int (*acceptable)(void *context, struct dentry *de),
+			void *context)
+{
+	struct fuse_inode_handle handle;
+	struct fuse_inode_handle parent;
+
+	if (fh_len < 3 || fileid_type > 2)
+		return NULL;
+
+	if (fileid_type == 2) {
+		if (fh_len < 6)
+			return NULL;
+
+		parent.nodeid = (u64) fh[3] << 32;
+		parent.nodeid |= (u64) fh[4];
+		parent.generation = fh[5];
+	} else {
+		parent.nodeid = 0;
+		parent.generation = 0;
+	}
+
+	handle.nodeid = (u64) fh[0] << 32;
+	handle.nodeid |= (u64) fh[1];
+	handle.generation = fh[2];
+
+	return fuse_export_operations.
+		find_exported_dentry(sb, &handle, &parent, acceptable, context);
+}
+
+static struct export_operations fuse_export_operations = {
+	.get_dentry	= fuse_get_dentry_old,
+	.encode_fh      = fuse_encode_fh,
+	.decode_fh	= fuse_decode_fh,
+};
+#endif
 #endif
 
 static struct super_operations fuse_super_operations = {
@@ -837,8 +940,12 @@ static decl_subsys(fs, NULL, NULL);
 static decl_subsys(fuse, NULL, NULL);
 static decl_subsys(connections, NULL, NULL);
 
+#ifdef KERNEL_2_6_24_PLUS
+static void fuse_inode_init_once(struct kmem_cache *cachep, void *foo)
+#else
 static void fuse_inode_init_once(void *foo, struct kmem_cache *cachep,
 				 unsigned long flags)
+#endif
 {
 	struct inode * inode = foo;
 
@@ -861,10 +968,18 @@ static int __init fuse_fs_init(void)
 	if (err)
 		goto out_unreg;
 
+#ifdef KERNEL_2_6_23_PLUS
+	fuse_inode_cachep = kmem_cache_create("fuse_inode",
+					      sizeof(struct fuse_inode),
+					      0, SLAB_HWCACHE_ALIGN,
+					      fuse_inode_init_once);
+#else
 	fuse_inode_cachep = kmem_cache_create("fuse_inode",
 					      sizeof(struct fuse_inode),
 					      0, SLAB_HWCACHE_ALIGN,
 					      fuse_inode_init_once, NULL);
+#endif
+
 	err = -ENOMEM;
 	if (!fuse_inode_cachep)
 		goto out_unreg2;
